@@ -2,7 +2,7 @@ package RedisDB;
 
 use strict;
 use warnings;
-our $VERSION = "2.00";
+our $VERSION = "2.09_01";
 $VERSION = eval $VERSION;
 
 use RedisDB::Error;
@@ -103,7 +103,9 @@ sub new {
     $self->{host} ||= 'localhost';
     $self->{_replies}       = [];
     $self->{_to_be_fetched} = 0;
-    $self->{database} ||= 0;
+    $self->{database}            ||= 0;
+    $self->{reconnect_attempts}  ||= 5;
+    $self->{reconnect_delay_max} ||= 8;
     $self->_init_parser;
     $self->_connect unless $self->{lazy};
     return $self;
@@ -154,6 +156,7 @@ sub _connect {
 
     $self->{_pid} = $$;
 
+    delete $self->{_socket};
     if ( $self->{path} ) {
         $self->{_socket} = IO::Socket::UNIX->new(
             Type => SOCK_STREAM,
@@ -161,12 +164,24 @@ sub _connect {
         ) or croak "Can't connect to redis server socket at `$self->{path}': $!";
     }
     else {
-        $self->{_socket} = IO::Socket::INET->new(
-            PeerAddr => $self->{host},
-            PeerPort => $self->{port},
-            Proto    => 'tcp',
-            ( $self->{timeout} ? ( Timeout => $self->{timeout} ) : () ),
-        ) or croak "Can't connect to redis server $self->{host}:$self->{port}: $!";
+        my $attempts = $self->{reconnect_attempts};
+        my $delay;
+        my $err;
+        while ( not $self->{_socket} and $attempts ) {
+            sleep $delay if $delay;
+            $self->{_socket} = IO::Socket::INET->new(
+                PeerAddr => $self->{host},
+                PeerPort => $self->{port},
+                Proto    => 'tcp',
+                ( $self->{timeout} ? ( Timeout => $self->{timeout} ) : () ),
+            ) or $err = $!;
+            $delay = $delay ? ( 1 + rand ) * $delay : 1;
+            $delay = $self->{reconnect_delay_max} if $delay > $self->{reconnect_delay_max};
+            $attempts--;
+        }
+        unless ( $self->{_socket} ) {
+            croak "Can't connect to redis server $self->{host}:$self->{port}: $err";
+        }
     }
 
     if ( $self->{timeout} ) {
@@ -246,7 +261,7 @@ sub _recv_data_nb {
             # if there are some replies lost
             confess "Server closed connection. Some data was lost."
               if $self->{_parser}->callbacks
-                  or $self->{_in_multi};
+              or $self->{_in_multi};
 
             # clean disconnect, try to reconnect
             $self->{warnings} and warn "Disconnected, trying to reconnect";
@@ -433,13 +448,13 @@ sub get_reply {
 
     croak "We are not waiting for reply"
       unless @{ $self->{_replies} }
-          or $self->{_to_be_fetched}
-          or $self->{_subscription_loop};
+      or $self->{_to_be_fetched}
+      or $self->{_subscription_loop};
     croak "You can't read reply in child process" unless $self->{_pid} == $$;
-    while ( not @{$self->{_replies}} ) {
+    while ( not @{ $self->{_replies} } ) {
         my $ret = $self->{_socket}->recv( my $buffer, 131072 );
         unless ( defined $ret ) {
-            next if $! == EINTR;
+            next if $! == EINTR or $! == 0;
             confess "Error reading reply from server: $!";
         }
         if ( $buffer ne '' ) {
@@ -640,7 +655,7 @@ answer.  Croaks in case of the error.
 
 sub shutdown {
     my $self = shift;
-    $self->send_command_cb('SHUTDOWN', @_);
+    $self->send_command_cb( 'SHUTDOWN', @_ );
     return;
 }
 
@@ -846,12 +861,12 @@ sub subscription_loop {
     croak "Already in subscription loop" if $self->{_subscribtion_loop};
     croak "You can't start subscription loop while in pipelining mode."
       if $self->{_parser}->callbacks
-          or @{ $self->{_replies} };
+      or @{ $self->{_replies} };
     $self->{_subscribed}        = {};
     $self->{_psubscribed}       = {};
     $self->{_subscription_cb}   = $args{default_callback};
     $self->{_subscription_loop} = 1;
-    $self->{_parser}->set_default_callback(\&_queue);
+    $self->{_parser}->set_default_callback( \&_queue );
 
     if ( $args{subscribe} ) {
         while ( my $channel = shift @{ $args{subscribe} } ) {
@@ -870,50 +885,50 @@ sub subscription_loop {
     croak "You must subscribe at least to one channel"
       unless ( keys %{ $self->{_subscribed} } or keys %{ $self->{_psubscribed} } );
 
-    while ( my $msg = $self->get_reply ) {
+    while ( $self->{_subscription_loop} ) {
+        my $msg = $self->get_reply;
         confess "Expected multi-bulk reply, but got $msg" unless ref $msg;
         if ( $msg->[0] eq 'message' ) {
-            $self->{_subscribed}{ $msg->[1] }( $self, $msg->[1], undef, $msg->[2] );
+            $self->{_subscribed}{ $msg->[1] }( $self, $msg->[1], undef, $msg->[2] )
+              if $self->{_subscribed}{ $msg->[1] };
         }
         elsif ( $msg->[0] eq 'pmessage' ) {
-            $self->{_psubscribed}{ $msg->[1] }( $self, $msg->[2], $msg->[1], $msg->[3] );
+            $self->{_psubscribed}{ $msg->[1] }( $self, $msg->[2], $msg->[1], $msg->[3] )
+              if $self->{_psubscribed}{ $msg->[1] };
         }
-        elsif ( $msg->[0] eq 'subscribe' or $msg->[0] eq 'psubscribe' ) {
+        elsif ( $msg->[0] =~ /^p?(un)?subscribe/ ) {
 
             # ignore
         }
-        elsif ( $msg->[0] eq 'unsubscribe' ) {
-            delete $self->{_subscribed}{ $msg->[1] };
-
-            # TODO think about it, not exactly correct
-            last unless $msg->[2] or %{ $self->{_psubscribed} };
-        }
-        elsif ( $msg->[0] eq 'punsubscribe' ) {
-            delete $self->{_psubscribed}{ $msg->[1] };
-            last unless $msg->[2] or %{ $self->{_subscribed} };
-        }
         else {
             confess "Got unknown reply $msg->[0] in subscription mode";
         }
     }
-    $self->{_to_be_fetched} = 0;
-    $self->_connect;
     return;
 }
 
 =head2 $self->subscribe($channel[, $callback])
 
-Subscribe to the additional I<$channel>. If I<$callback> is not specified,
-default callback will be used.
+Subscribe to the I<$channel>. If I<$callback> is not specified, default
+callback will be used. If you are invoking I<subscribe> outside of subscription
+loop, I<$callback> is ignored.
 
 =cut
 
 sub subscribe {
     my ( $self, $channel, $callback ) = @_;
-    croak "Must be in subscription loop to subscribe" unless $self->{_subscription_loop};
+    unless ( $self->{_subscription_loop} ) {
+        $self->{_subscription_loop} = -1;
+        $self->{_parser}->set_default_callback( \&_queue );
+    }
     croak "Subscribe to what channel?" unless length $channel;
-    $callback ||= $self->{_subscription_cb}
-      or croak "Callback for $channel not specified, neither default callback defined";
+    if ( $self->{_subscription_loop} > 0 ) {
+        $callback ||= $self->{_subscription_cb}
+          or croak "Callback for $channel not specified, neither default callback defined";
+    }
+    else {
+        $callback = 1;
+    }
     $self->{_subscribed}{$channel} = $callback;
     $self->send_command( "SUBSCRIBE", $channel );
     return;
@@ -921,17 +936,26 @@ sub subscribe {
 
 =head2 $self->psubscribe($pattern[, $callback])
 
-Subscribe to additional channels matching I<$pattern>. If I<$callback> is not specified, default
-callback will be used.
+Subscribe to channels matching I<$pattern>. If I<$callback> is not specified,
+default callback will be used. If you are invoking I<psubscribe> outside of
+subscription loop, I<$callback> is ignored.
 
 =cut
 
 sub psubscribe {
     my ( $self, $channel, $callback ) = @_;
-    croak "Must be in subscription loop to subscribe" unless $self->{_subscription_loop};
+    unless ( $self->{_subscription_loop} ) {
+        $self->{_subscription_loop} = -1;
+        $self->{_parser}->set_default_callback( \&_queue );
+    }
     croak "Subscribe to what channel?" unless length $channel;
-    $callback ||= $self->{_subscription_cb}
-      or croak "Callback for $channel not specified, neither default callback defined";
+    if ( $self->{_subscription_loop} > 0 ) {
+        $callback ||= $self->{_subscription_cb}
+          or croak "Callback for $channel not specified, neither default callback defined";
+    }
+    else {
+        $callback = 1;
+    }
     $self->{_psubscribed}{$channel} = $callback;
     $self->send_command( "PSUBSCRIBE", $channel );
     return;
@@ -940,25 +964,55 @@ sub psubscribe {
 =head2 $self->unsubscribe([@channels])
 
 Unsubscribe from the listed I<@channels>. If no channels was specified,
-unsubscribe from all the channels.
+unsubscribe from all the channels to which you have subscribed using
+I<subscribe>.
 
 =cut
 
 sub unsubscribe {
     my $self = shift;
-    return $self->send_command( "UNSUBSCRIBE", @_ );
+    if (@_) {
+        delete $self->{_subscribed}{$_} for @_;
+    }
+    else {
+        $self->{_subscribed} = {};
+    }
+    if ( %{ $self->{_subscribed} } or %{ $self->{_psubscribed} } ) {
+        return $self->send_command( "UNSUBSCRIBE", @_ );
+    }
+    else {
+        delete $self->{_subscription_loop};
+        $self->{_to_be_fetched} = 0;
+        return $self->_connect;
+    }
 }
 
 =head2 $self->punsubscribe([@patterns])
 
 Unsubscribe from the listed I<@patterns>. If no patterns was specified,
-unsubscribe from all the channels to which you subscribed using I<psubscribe>.
+unsubscribe from all the channels to which you have subscribed using
+I<psubscribe>.
 
 =cut
 
 sub punsubscribe {
     my $self = shift;
-    return $self->send_command( "PUNSUBSCRIBE", @_ );
+    if (@_) {
+        delete $self->{_psubscribed}{$_} for @_;
+    }
+    else {
+        $self->{_psubscribed} = {};
+    }
+    if (   %{ $self->{_subscribed} }
+        or %{ $self->{_psubscribed} } )
+    {
+        return $self->send_command( "PUNSUBSCRIBE", @_ );
+    }
+    else {
+        delete $self->{_subscription_loop};
+        $self->{_to_be_fetched} = 0;
+        return $self->_connect;
+    }
 }
 
 =head2 $self->subscribed
